package ru.nsu.fit.mustakov.graphics;

import java.awt.*;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

public class FunctionPlot extends Plot {

    private final static double DEFAULT_SECTIONS_PER_PIXEL = 0.05;
    private double sectionsPerPixel = DEFAULT_SECTIONS_PER_PIXEL;
    
    private int MIN_SECTION_DRAW_STEP = 10;
    
    private MathFunction function;

    private final static Color DEFAULT_FUNCTION_COLOR = Color.RED;
    private final static Color DEFAULT_SECTION_COLOR = Color.BLACK;
    private final static Color DEFAULT_NUMBER_COLOR = Color.BLACK;
    
    private Color functionColor;
    private Color sectionColor;
    private Color numberColor;
    
    public FunctionPlot() {
        this(null);
    }

    public FunctionPlot(MathFunction mathFunction) {
        super();
        function = mathFunction;
        functionColor = DEFAULT_FUNCTION_COLOR;
        sectionColor = DEFAULT_SECTION_COLOR;
        numberColor = DEFAULT_NUMBER_COLOR;
    }

    public void setFunction(MathFunction function) {
        this.function = function;
        dispatchChangeEvent();
    }

    public MathFunction getFunction() {
        return function;
    }

    @Override
    public void draw(Graphics2D g2, Rectangle2D area) {
        g2.drawLine((int)area.getMinX(), (int)area.getCenterY(), (int)area.getMaxX(), (int)area.getCenterY());
        g2.drawLine((int)area.getCenterX(), (int)area.getMinY(), (int)area.getCenterX(), (int)area.getMaxY());
        drawSections(g2, area);
        drawFunction(g2, area);
    }
    
    public void setFunctionColor(Color color) {
        if (color == null) {
            this.functionColor = DEFAULT_FUNCTION_COLOR;
        }
        else {
            this.functionColor = color;
        }
    }
    
    public void setSectionColor(Color color) {
        if (color == null) {
            this.sectionColor = DEFAULT_SECTION_COLOR;
        }
        else {
            this.sectionColor = color;
        }
    }
    
    public void setNumberColor(Color color) {
        if (color == null) {
            this.numberColor = DEFAULT_NUMBER_COLOR;
        }
        else {
            this.numberColor = color;
        }
    }
    
    @Override
    public void zoom(double percent) {
        sectionsPerPixel = DEFAULT_SECTIONS_PER_PIXEL * percent;
        dispatchChangeEvent();
    }

    public double getZoom() {
        return sectionsPerPixel / DEFAULT_SECTIONS_PER_PIXEL;
    }
    
    private void drawFunction(Graphics2D g2, Rectangle2D area) {
        if (function != null) {
            Color oldColor = g2.getColor();
            g2.setColor(functionColor);
            Point2D prevPoint = null;
            for (double x = area.getMinX(); x < area.getMaxX(); ++x) {
                double realX = x - area.getCenterX();
                double y = function.calculate(realX * sectionsPerPixel) / sectionsPerPixel;
                if (!Double.isNaN(y)) {
                    y = -y + area.getCenterY();
                    if (y <= area.getMaxY() && y >= area.getMinY()) {
                        if (prevPoint != null) {
                            GraphicsUtilities.drawLine(g2, prevPoint, new Point.Double(x,y));
                        }
                        else {
                            g2.draw(new Pixel(x, y));
                        }
                        prevPoint = new Point.Double(x, y);
                    }
                }
                else {
                    prevPoint = null;
                }
            }
            g2.setColor(oldColor);
        }
    }
    
    private void drawSections(Graphics2D g2, Rectangle2D area) {
        final int SECTION_LINE_LENGTH = 6;
        
        double firstSection = MIN_SECTION_DRAW_STEP * sectionsPerPixel;
        double step = 1;
        int power = 0;
        while (firstSection < step/10) {
            step /= 10;
            --power;
        }
        while (firstSection > step) {
            step *= 10;
            ++power;
        }
        double stepPixel = toPixel(step);
        for (double x = stepPixel; x < area.getMaxX(); x += stepPixel) {
            Point2D first = new Point.Double(-x, SECTION_LINE_LENGTH / 2);
            first = toMonitorPoint(first, area);
            Point2D second = new Point.Double(-x, - SECTION_LINE_LENGTH / 2);
            second = toMonitorPoint(first, area);
            GraphicsUtilities.drawLine(g2, first, second);
            
            first = new Point.Double(x, SECTION_LINE_LENGTH / 2);
            first = toMonitorPoint(first, area);
            second = new Point.Double(x, - SECTION_LINE_LENGTH / 2);
            second = toMonitorPoint(first, area);
            GraphicsUtilities.drawLine(g2, first, second);
            
            first = new Point.Double(-SECTION_LINE_LENGTH / 2, x);
            first = toMonitorPoint(first, area);
            second = new Point.Double(SECTION_LINE_LENGTH / 2, x);
            second = toMonitorPoint(first, area);
            GraphicsUtilities.drawLine(g2, first, second);
            
            first = new Point.Double(-SECTION_LINE_LENGTH / 2, -x);
            first = toMonitorPoint(first, area);
            second = new Point.Double(SECTION_LINE_LENGTH / 2, -x);
            second = toMonitorPoint(first, area);
            GraphicsUtilities.drawLine(g2, first, second);
        }
    }
    
    private double toPixel(double value) {
        return value / sectionsPerPixel;
    }
    
    private Point2D toMonitorPoint(Point2D plotPoint, Rectangle2D area) {
        return new Point.Double(plotPoint.getX() + area.getCenterX(),
                                -plotPoint.getY() + area.getCenterY());
    }
}
